# ROTOR.py
# Managing only one rotor

import string
import sys

N = 26
ALPHA = string.ascii_uppercase

class Rotor:
    def __init__(self,PI,trigo=False,inverse=False):
        self.PI = Rotor.rotor2int(PI)
        self.PI_INV = Rotor.invrot(self.PI)
        self.PI_SHIFT = Rotor.calculate_shift(self.PI)
        self.PI_SHIFT_INV = Rotor.cal_shift_inv(self.PI)
        self.PI_MIR = Rotor.cal_miroir(self.PI)
        self.SHIFT_MIR = Rotor.calculate_shift(self.PI_MIR)
        self.SHIFT_MIR_INV = Rotor.cal_shift_inv(self.PI_MIR)
        self.TRIGO = trigo
        self.INVERSE = inverse
        self.key = None
        self.rewind()
        if not self.INVERSE:
            self.SHIFT = self.PI_SHIFT
            self.SHIFT_INV = self.PI_SHIFT_INV
            self.EDGE = ALPHA
        else:
            self.SHIFT = self.SHIFT_MIR
            self.SHIFT_INV = self.SHIFT_MIR_INV
            tb = []
            for i in range(N,0,-1):
                tb.append( ALPHA[ i%N ] )
            self.EDGE = "".join(tb)

    def set_key(self, key):
        self.key = key

    def get_key(self):
        return self.key

    def get_external_key(self):
        return self.EDGE[ self.key ]

    def set_external_key(self, lettre):
        idx = (self.EDGE).index(lettre)
        self.key = idx

    def zero(self):
        self.key = 0

    def rewind(self):
        if self.TRIGO:
            self.key = 1
        else:
            self.key = 25

    def info(self):
        ch = ""
        ch += "Key: " +  ALPHA[self.key] + "\n"
        ch += "PI: " +  str(self.PI) + "\n"
        ch += "PI reverse: " + str(self.PI_INV) + "\n"
        ch += "PI mirror : " + str(self.PI_MIR) + "\n"
        ch += "Shift: " + str(self.PI_SHIFT) + "\n"
        ch += "Shift mirror: " + str(self.PI_SHIFT_INV) + "\n"
        ch += "Shift reverse: " + str(self.SHIFT_MIR) + "\n"
        ch += "Shift mirror reverse: " + str(self.SHIFT_MIR_INV) + "\n"
        if self.TRIGO : direction = "Trigonometric"
        else: direction = "Clockwise"
        ch += "Direction: " + direction + "\n"
        ch += "Rotor Reverse: " + str(self.INVERSE) + "\n"
        ch += "Edge : " + self.EDGE + "\n"
        return ch

    def move_forward(self):
        self.key = (self.key + 1)%N

    def move_backward(self):
        self.key = (self.key - 1)%N

    def cipher_a_char(self, car):
        return (car + self.SHIFT[ (car + self.key)%N ]) % N

    def decipher_a_char(self, car):
        return (car + self.SHIFT_INV[ (car + self.key)%N ]) % N

    #-------- Methodes statiques ------------------
    def rotor2int(rotor):
        tb = [0]*N
        for i in range(N):
            c = rotor[i].upper()
            tb[i] = ALPHA.index(c)
        return tb

    def rotor2char(rotor):
        ch = ""
        for i in range(N):
            ch += ALPHA[rotor[i]]
        return ch

    def decal2pi(decalage):
        pi = [0]*N
        for i in range(N):
            pi[i] = (i + decalage[i])%N
        return pi

    def invrot(rotor):
        """ inverse un rotor """
        PIinv = [0]*N
        for x in range(N):
            y = rotor[x]
            PIinv[y] = (rotor[x] - (y - x))%N
        return PIinv

    def calculate_shift(pi):
        Decalage = [0] * N
        for i in range(N):
            j = pi[i]
            Decalage[i] = ( j - i ) % N
        return Decalage

    def cal_shift_inv(pi):
        Decal_inv = [0] * N
        tampon = [0] * N
        for i in range(N):
            j = pi[i]
            tampon[j] = i
        for i in range(N):
            Decal_inv[i] = (tampon[i] - i) % N
        return Decal_inv

    def cal_miroir(pi):
        PI_mir = [0]*N
        pi_inv = Rotor.invrot(pi)
        for i in range(N):
            x = (N - i) % N
            y = (N - pi_inv[i]) % N
            PI_mir[x] = y 
        return PI_mir

    def valide(pi):
        flag = [False]*N
        if len(pi) < N:
            return False
        for i in range(N):
            if not( pi[i] in ALPHA ):
                return False
            idx = ALPHA.index(pi[i])
            if flag[idx]:
                return False
            flag[idx] = True
        return True 
            
#-------------------------------------------
def chiffre(hebern, plain):
    crypto = ""
    for car in plain:
        #print("%2d " % (c.key), end=": ")
        x = ALPHA.index(car)
        if hebern.TRIGO:
            hebern.move_backward()
        else:
            hebern.move_forward()
        y = hebern.cipher_a_char(x)
        sortie = ALPHA[y]
        crypto += sortie
        #print("%c -> %c" % (car,sortie), end=",")
    return crypto

def dechiffre(hebern, plain):
    crypto = ""
    for car in plain:
        x = ALPHA.index(car)
        if hebern.TRIGO:
            hebern.move_backward()
        else:
            hebern.move_forward()
        y = hebern.decipher_a_char(x)
        sortie = ALPHA[y]
        crypto += sortie
    return crypto

#-------------------------------------------
if __name__ == "__main__":    
    c = Rotor("OBCLJATDIUZXSWRHNYEVMGKFQP")
    plain = "ILFAITBEAU"
    print("Info : "); print(c.info())
    print("Crypto (clockwise, cipher)  : ", chiffre(c, plain))

    c.rewind()
    plain = "IRBIOLXQAC"
    print("Crypto (clockwise, decipher): ", dechiffre(c, plain))

    c.TRIGO = True
    plain = "ILFAITBEAU"
    c.rewind()
    print("Crypto: (trigo, cipher):  ", chiffre(c, plain))

    c.rewind()
    plain = "IANINWMMMG"
    print("Crypto: (trigo, decipher): ", dechiffre(c, plain))

    c = Rotor("FTQJVAXMWDSNHLRUCOKBPEIGZY")
    print(c.info())

    plain = "SENDMOREAMMUNITION"
    print("=======> plain: ", plain)
    c = Rotor("FTQJVAXMWDSNHLRUCOKBPEIGZY", trigo=False)
    c.set_external_key("D")
    print("Info : "); print(c.info())

    crypto = ""
    for car in plain:
        print("%2d " % (c.key), end=": ")
        x = ALPHA.index(car)
        c.move_forward()
        y = c.cipher_a_char(x)
        sortie = ALPHA[y]
        crypto += sortie
        print("%c -> %c" % (car,sortie), end=",")
    print()
    print(crypto)
    c = Rotor("FTQJVAXMWDSNHLRUCOKBPEIGZY", trigo=False, inverse=True)
    print("Info : "); print(c.info())

