# search_rotor.py
# recherche le rotor rapide
# on donne le cablage du rotor et le fichier cryptogramme

import ROTOR 
import sys
import string

ALPHA = string.ascii_uppercase

#======================================
def lit_rotor(nom):
    f = open("ROTORS/" + nom + ".rot")
    permutation = f.readline()
    f.close()
    return permutation.strip()
#======================================
def lit_stdin():
    entree = ""
    leTexte = sys.stdin.readlines()
    for i in range(len(leTexte)):
        entree += leTexte[i]
    flux_entre = ""
    for car in entree:
        if car.upper() in ALPHA:
            flux_entre += car.upper()
    return flux_entre
#======================================
def ic(texte):
    lg = len(texte)
    if lg < 2:
        return 0.038
    freq = [0] * 26
    for i in range(lg):
        idx = ALPHA.find(texte[i])
        freq[ idx ] += 1
    total = 0
    for i in range(26):
        if freq[i] >= 2:
            total += freq[ i ] * (freq[i] - 1)
    return total / (lg * (lg-1))
#======================================

# 1 - Initialisations (rotor, crypto)
fast_rotor = lit_rotor(sys.argv[1])
crypto = lit_stdin()
c = ROTOR.Rotor(fast_rotor)

# 2 - On boucle: on dechiffre le message par
#     tranche de 26 lettres et pour chaque
#     tranche on calcule l'IC et en final
#     on fait une moyenne. A chaque tour
#     de boucle, on change la position initiale
#     du rotor testé.
ic_max = 0
cle = 0
print("====> Find for the good rotor")
print("Starting position  Average IC")
for i in range(26):
    ic_total = 0
    c.set_key(i)
    k = 0
    for j in range(0,len(crypto),26):
        tranche = crypto[j:j+26]
        #print(tranche)
        mono_substi = ROTOR.dechiffre(c, tranche)
        ic_tranche = ic(mono_substi)
        if len(tranche) < 26:
            ic_tranche = ic_tranche * (len(tranche)/26)
        #print(tranche, mono_substi, ic_tranche)
        ic_total += ic_tranche
        k += 1
    if ic_total/k > ic_max:
        ic_max = ic_total/k
        cle = ALPHA[i]
    print("   %c:              %5.4f" % (ALPHA[i], ic_total/k))
print("IC max: %5.4f, key: %c" % (ic_max, cle))

#print( ROTOR.dechiffre(c, crypto))
#c.set_key(ALPHA.find(cle))
#print(ROTOR.dechiffre(c,crypto[0:26]))

ic_max = 0
wheel = 0
print("==> If the rotor is correct, search for ratchet right wheel position")
print("Starting position       Average IC")
for i in range(26):
    #print("==> ratchet wheel left:", i)
    debut = 0
    fin = i + 1
    c.set_key(ALPHA.find(cle))
    k = 0
    ic_total = 0
    while 1:
        tranche = crypto[debut:fin]
        mono_substi = ROTOR.dechiffre(c, tranche)
        ic_tranche = ic(mono_substi)
        #if len(tranche) < 26:
        #    ic_tranche = ic_tranche * (len(tranche)/26)
        print("%3d %3d -> %30s, %30s, %5.4f" % \
            (debut, fin, tranche, mono_substi, ic_tranche))
        if fin > len(crypto): break
        debut = fin
        fin = debut + 26
        k += 1
        ic_total += ic_tranche
    if ic_total/k > ic_max:
        ic_max = ic_total/k
        wheel = i
    print("     %2d                 %4.4f" % (i, ic_total / k))
ratchet_wheel = (13 - (wheel+1)) % 26
print("---> ic max: %5.4f, start of ratchet wheel: %c" % \
    (ic_max, ALPHA[ratchet_wheel]))

